from trubrics.feedback import Feedback as collect
from trubrics.feedback import save_to_trubrics as save
from trubrics.trubrics_platform import init

try:
    from importlib.metadata import version  # type: ignore
except ImportError:
    # for python<3.8
    from importlib_metadata import version  # type: ignore

__version__ = version("trubrics")

__all__ = ["collect", "init", "save"]



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field


class Response(BaseModel):
    type: str
    score: Optional[str] = None
    text: Optional[str] = None


class Feedback(BaseModel):
    """
    The Feedback object represents all data contained as a feedback response from a user.

    Example:
        ```py
        import trubrics

        feedback = trubrics.collect(
            component_name="default",
            model="default_model",
            response={
                "type": "thumbs",
                "score": "👎",
                "text": "A comment / textual feedback from the user."
            },
            user_id="user_id",
            tags=["satisfaction"],
            metadata={"prompt": "What is the meaning of user feedback for AI?"}
        )
        ```

    Attributes:
        component_name: the name of the component that the feedback response is saved to
        model: the model name / version
        response: the user response, with a type, score and text
        created_on: datetime of response (in UTC)
        user_id: an optional user id
        tags: optional tags
        metadata: optional metadata, such as model prompts & predictions
    """

    component_name: str
    model: str
    response: Response
    created_on: datetime = Field(default_factory=datetime.utcnow)
    user_id: Optional[str] = None
    tags: list = []
    metadata: dict = {}



from pathlib import Path

import typer
from rich import print as rprint

from trubrics.cli.run import validate_trubric_run_context
from trubrics.trubrics_platform.trubrics_config import TrubricsDefaults

app = typer.Typer(pretty_exceptions_show_locals=False)


def version_callback(value: bool):
    if value:
        import trubrics

        typer.echo(trubrics.__version__)
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=True),
):
    return None


@app.command()
def run(
    save_ui: bool = typer.Option(False, prompt="Would you like to save your trubric to the UI?"),
    run_context_path: str = typer.Option(
        default="example", prompt="Enter the path to your trubric run .py file. Press enter for an"
    ),
    trubric_output_file_path: str = "./my_new_trubric.json",
    raise_on_failure: bool = True,
):
    """Runs an example trubric (list of model validations) on the titanic dataset.

    Args:
        save_ui: whether to save validations to the UI with in app user authentication
        run_context_path: path to the trubrics run context
        trubric_output_file_path: path to save your output trubric file
        raise_on_failure: raise an exception on failure of trubric
    """
    trubric_run_path = None
    if run_context_path != "example":
        trubric_run_path = Path(run_context_path).absolute()
        if not trubric_run_path.exists():
            rprint(f"[red]Path '{trubric_run_path}' not found.[red]")
            raise typer.Abort()
        rc = validate_trubric_run_context(str(trubric_run_path)).RUN_CONTEXT
    else:
        from trubrics.example.trubric_run import RUN_CONTEXT as rc  # type: ignore
    rprint(
        f"\nRunning trubric from file '{trubric_run_path or 'trubrics.example.trubric_run'}' with model"
        f" '{rc.model_name}' and dataset '{rc.data_context.name}'.\n"
    )
    new_trubric = rc.set_new_trubric()
    if save_ui:
        raise NotImplementedError(
            "It is currently not possible to save validations to Trubrics. Get in touch with us if this is an issue."
        )
    else:
        defaults = TrubricsDefaults()
        new_trubric.save_local(trubric_output_file_path, raise_on_failure=raise_on_failure)
        rprint(
            "\n[bold orange_red1]Be sure to check out our docs to see how you can leverage the Trubrics platform."
            f"\n\n{defaults.demo_sign_up_url}[bold orange_red1]\n"
        )


if __name__ == "__main__":
    app()



"""
File of HTTP requests to Firestore Rest API.
"""
import json
from datetime import datetime

import requests  # type: ignore


def dict_to_firestore_document(python_dict):
    firestore_compatible = {"fields": {}}
    for key, value in python_dict.items():
        if value is None:
            firestore_compatible["fields"][key] = {"nullValue": value}
        elif isinstance(value, str):
            firestore_compatible["fields"][key] = {"stringValue": value}
        elif isinstance(value, bool):
            firestore_compatible["fields"][key] = {"booleanValue": value}
        elif isinstance(value, int):
            firestore_compatible["fields"][key] = {"integerValue": value}
        elif isinstance(value, float):
            firestore_compatible["fields"][key] = {"doubleValue": value}
        elif isinstance(value, datetime):
            firestore_compatible["fields"][key] = {"timestampValue": value.isoformat() + "Z"}
        elif isinstance(value, dict):
            firestore_compatible["fields"][key] = {"mapValue": dict_to_firestore_document(value)}
        elif isinstance(value, list):
            array_values = []
            for item in value:
                if value is None:
                    array_values.append({"nullValue": item})
                elif isinstance(item, str):
                    array_values.append({"stringValue": item})
                elif isinstance(item, bool):
                    array_values.append({"booleanValue": item})
                elif isinstance(item, int):
                    array_values.append({"integerValue": item})
                elif isinstance(item, float):
                    array_values.append({"doubleValue": item})
                elif isinstance(item, datetime):
                    array_values.append({"timestampValue": item.isoformat() + "Z"})
                elif isinstance(item, dict):
                    array_values.append({"mapValue": dict_to_firestore_document(item)})
            firestore_compatible["fields"][key] = {"arrayValue": {"values": array_values}}
    return firestore_compatible


def get_trubrics_firestore_api_url(auth, project_id):
    structured_query = {
        "structuredQuery": {
            "from": [{"collectionId": "organisations"}],
            "where": {
                "fieldFilter": {
                    "field": {"fieldPath": "users"},
                    "op": "ARRAY_CONTAINS",
                    "value": {
                        "stringValue": auth["email"],
                    },
                }
            },
        }
    }
    organisation_route = json.loads(
        requests.post(
            f"https://firestore.googleapis.com/v1/projects/{project_id}/databases/(default)/documents:runQuery",
            headers={"Content-Type": "application/json", "Authorization": f"Bearer {auth['idToken']}"},
            data=json.dumps(structured_query),
        ).text
    )[0]["document"]["name"]
    return f"https://firestore.googleapis.com/v1/{organisation_route}"


def list_components_in_organisation(firestore_api_url, auth):
    r = requests.get(
        firestore_api_url + "/feedback" + "?pageSize=50",
        headers={"Content-Type": "application/json", "Authorization": f"Bearer {auth['idToken']}"},
    )
    r.raise_for_status()
    components_res = json.loads(r.text)

    all_components = []
    if len(components_res) != 0:
        for component in components_res["documents"]:
            if component.get("fields", {}).get("archived", {}).get("booleanValue", {}) is False:
                all_components.append(component["name"].split("/")[-1])
    return all_components


def record_feedback(auth, firestore_api_url, document_dict):
    url = firestore_api_url + f"/feedback/{document_dict['component_name']}/responses"
    res = json.loads(
        requests.post(
            url,
            headers={"Content-Type": "application/json", "Authorization": f"Bearer {auth['idToken']}"},
            data=json.dumps(dict_to_firestore_document(document_dict)),
        ).text
    )
    return res


def add_document_to_project_subcollection(auth, firestore_api_url, project, subcollection, document_id, document_dict):
    url = firestore_api_url + f"/projects/{project}/{subcollection}/?documentId={document_id}"
    res = json.loads(
        requests.post(
            url,
            headers={"Content-Type": "application/json", "Authorization": f"Bearer {auth['idToken']}"},
            data=json.dumps(dict_to_firestore_document(document_dict)),
        ).text
    )
    return res


def list_projects_in_organisation(firestore_api_url, auth):
    r = requests.get(
        firestore_api_url + "/projects" + "?pageSize=50",
        headers={"Content-Type": "application/json", "Authorization": f"Bearer {auth['idToken']}"},
    )
    r.raise_for_status()
    projects_res = json.loads(r.text)

    all_projects = []
    if len(projects_res) != 0:
        for project in projects_res["documents"]:
            if project.get("fields", {}).get("archived", {}).get("booleanValue", {}) is False:
                all_projects.append(project["name"].split("/")[-1])
    return all_projects



from datetime import datetime
from typing import Optional

import streamlit as st

from trubrics.feedback import config
from trubrics.feedback.dataclass import Feedback, Response
from trubrics.feedback.save_to_trubrics import save_to_trubrics as save
from trubrics.trubrics_platform.auth import init


class FeedbackCollector:
    def __init__(
        self,
        component_name: str,
        email: Optional[str],
        password: Optional[str],
        firebase_api_key: Optional[str] = None,
        firebase_project_id: Optional[str] = None,
    ):
        """
        Args:
            component_name: the component name that has been created in Trubrics
            email: a Trubrics account email
            password: a Trubrics account password
        """
        self.component_name = component_name
        if email and password:
            self.trubrics_config = init(
                email=email,
                password=password,
                firebase_api_key=firebase_api_key,
                firebase_project_id=firebase_project_id,
            )

    def st_feedback(
        self,
        feedback_type: str,
        model: str,
        tags: list = [],
        metadata: dict = {},
        response: Optional[Response] = None,
        user_id: Optional[str] = None,
        created_on: Optional[datetime] = None,
        key: Optional[str] = None,
        open_feedback_label: Optional[str] = None,
        save_to_trubrics: bool = True,
    ) -> Optional[dict]:
        """
        Collect ML model user feedback with UI components from a Streamlit app.

        Args:
            feedback_type: type of feedback to be collected

                - textbox: open textbox feedback
                - thumbs: 👍 / 👎 UI buttons
                - faces: 😞 / 🙁 / 😐 / 🙂 / 😀 UI buttons
            model: the model used - can be a model version, a link to the saved model artifact in cloud storage, etc
            tags: a list of tags for the feedback
            metadata: any data to save with the feedback
            response: a dict of user responses to save with the feedback for feedback_type='custom'
            user_id: an optional reference to a user, for example a username if there is a login, a cookie ID, etc
            key: a key for the streamlit components (necessary if calling this method multiple times with the same type)
            open_feedback_label: label of optional text_input for "faces" or "thumbs" feedback_type
            save_to_trubrics: whether to save the feedback to Trubrics, or just to return the feedback object
        """
        if key is None:
            key = feedback_type
        if feedback_type == "textbox":
            if response:
                raise ValueError("For feedback_type='textbox', user_response must be None.")
            text = self.st_textbox_ui(key, label=open_feedback_label)
            if text:
                response = Response(type=feedback_type, score=None, text=text)
                return self._save_feedback(
                    model=model,
                    response=response,
                    user_id=user_id,
                    tags=tags,
                    metadata=metadata,
                    created_on=created_on,
                    save_to_trubrics=save_to_trubrics,
                )
        elif feedback_type in ("thumbs", "faces"):
            if response:
                raise ValueError(
                    f"For feedback_type='{feedback_type}', response is set inside the component (must be None)."
                )
            return self._save_quantitative_feedback(
                model=model,
                feedback_type=feedback_type,
                user_id=user_id,
                metadata=metadata,
                tags=tags,
                created_on=created_on,
                key=key,
                open_feedback_label=open_feedback_label,
                save_to_trubrics=save_to_trubrics,
            )
        elif feedback_type == "custom":
            raise NotImplementedError(
                "This is currently not implemented. Get in touch to understand how to save custom feedback."
            )
            # if response:
            #     return self._save_feedback(...)
            # else:
            #     raise ValueError("For feedback_type='custom', title and description parameters must be specified.")
        else:
            raise ValueError("feedback_type must be one of ['textbox', 'faces', 'thumbs', 'custom'].")
        return None

    def _save_feedback(
        self,
        model: str,
        response: Response,
        user_id: Optional[str] = None,
        tags: list = [],
        metadata: dict = {},
        created_on: Optional[datetime] = None,
        save_to_trubrics: bool = True,
    ) -> Optional[dict]:
        feedback = Feedback(
            component_name=self.component_name,
            response=response,
            model=model,
            metadata=metadata,
            tags=tags,
            user_id=user_id,
        )
        if created_on:
            feedback.created_on = created_on
        if save_to_trubrics:
            res = save(trubrics_config=self.trubrics_config, feedback=feedback)
            if "error" in res:
                error_msg = f"Error in pushing feedback issue to Trubrics: {res}"
                st.error(error_msg)
            else:
                st.success(config.PLATFORM_SAVE)
        return feedback.dict()

    def _save_quantitative_feedback(
        self,
        feedback_type: str,
        key: str,
        open_feedback_label: Optional[str],
        model: str,
        user_id: Optional[str] = None,
        tags: list = [],
        metadata: dict = {},
        created_on: Optional[datetime] = None,
        save_to_trubrics: bool = True,
    ) -> Optional[dict]:
        if f"{key}_state" not in st.session_state:
            st.session_state[f"{key}_state"] = ""
        if f"{key}_save_button" not in st.session_state:
            st.session_state[f"{key}_save_button"] = False
        if f"previous_{key}_state" not in st.session_state:
            st.session_state[f"previous_{key}_state"] = ""

        def feedback_handler():
            response = Response(
                type=feedback_type,
                score=st.session_state[f"{key}_state"],
                text=st.session_state[f"{feedback_type}_open_feedback_{key}"].rstrip(),
            )
            st.session_state[f"previous_{key}_state"] = response
            st.session_state[f"{key}_state"] = ""

            self._enable_all_buttons(feedback_type=feedback_type)

        ui_state = getattr(self, f"st_{feedback_type}_ui")(
            key=key, disable_on_click=True if open_feedback_label else False
        )
        if ui_state:
            st.session_state[f"{key}_state"] = ui_state

        if open_feedback_label:
            if st.session_state.get(f"{key}_state_clicked", "") != "":
                st.text_input(open_feedback_label, key=f"{feedback_type}_open_feedback_{key}")
                st.button(
                    config.FEEDBACK_SAVE_BUTTON,
                    on_click=feedback_handler,
                    key=f"{key}_save_button",
                )
        else:
            if ui_state:
                response = Response(
                    type=feedback_type,
                    score=ui_state,
                    text=None,
                )
                return self._save_feedback(
                    model=model,
                    response=response,
                    user_id=user_id,
                    tags=tags,
                    metadata=metadata,
                    created_on=created_on,
                    save_to_trubrics=save_to_trubrics,
                )
        if st.session_state[f"{key}_save_button"]:
            return self._save_feedback(
                model=model,
                response=st.session_state[f"previous_{key}_state"],
                user_id=user_id,
                tags=tags,
                metadata=metadata,
                created_on=created_on,
                save_to_trubrics=save_to_trubrics,
            )
        return None

    @staticmethod
    def st_textbox_ui(key: Optional[str] = None, label: Optional[str] = None) -> Optional[str]:
        """
        Trubrics 'textbox' UI component.

        Args:
            key: a key for the streamlit components (necessary if calling this method multiple times with the same type)
            label: the textbox's streamlit label
        """
        if key is None:
            key = "textbox"

        if f"{key}_save_button" not in st.session_state:
            st.session_state[f"{key}_save_button"] = False

        if f"previous_{key}_state" not in st.session_state:
            st.session_state[f"previous_{key}_state"] = ""

        def clear_session_state():
            st.session_state[f"previous_{key}_state"] = st.session_state[f"{key}_title"]
            st.session_state[f"{key}_title"] = ""

        title = st.text_input(
            label=label or "Provide some feedback",
            key=f"{key}_title",
        )
        if title:
            st.button(config.FEEDBACK_SAVE_BUTTON, on_click=clear_session_state, key=f"{key}_save_button")
        if st.session_state[f"{key}_save_button"]:
            return st.session_state[f"previous_{key}_state"]
        else:
            return None

    def st_thumbs_ui(self, disable_on_click: bool = False, key: Optional[str] = None) -> Optional[str]:
        """
        Trubrics 'thumbs' UI component - two thumbs buttons.

        Args:
            disable_on_click: disable all other buttons when a button is clicked
            key: a key for the streamlit components (necessary if calling this method multiple times with the same type)
        """
        if key is None:
            key = "thumbs"

        button_states = [f"{key}_1", f"{key}_2"]
        col1, col2 = st.columns([1, 15])
        with col1:
            up = self._emoji_button("👍", key, disable_on_click, button_states, 1)
        with col2:
            down = self._emoji_button("👎", key, disable_on_click, button_states, 2)

        if up:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 1, "👍")
        elif down:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 2, "👎")
        else:
            return None

    def st_faces_ui(self, disable_on_click: bool = False, key: Optional[str] = None) -> Optional[str]:
        """
        Trubrics 'faces' UI component - two thumbs buttons.

        Args:
            disable_on_click: disable all other buttons when a button is clicked
            key: a key for the streamlit components (necessary if calling this method multiple times with the same type)
        """
        if key is None:
            key = "faces"

        button_states = [f"{key}_1", f"{key}_2", f"{key}_3", f"{key}_4", f"{key}_5"]
        col1, col2, col3, col4, col5 = st.columns([1, 1, 1, 1, 10])
        with col1:
            one = self._emoji_button("😞", key, disable_on_click, button_states, 1)
        with col2:
            two = self._emoji_button("🙁", key, disable_on_click, button_states, 2)
        with col3:
            three = self._emoji_button("😐", key, disable_on_click, button_states, 3)
        with col4:
            four = self._emoji_button("🙂", key, disable_on_click, button_states, 4)
        with col5:
            five = self._emoji_button("😀", key, disable_on_click, button_states, 5)

        if one:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 1, "😞")
        elif two:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 2, "🙁")
        elif three:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 3, "😐")
        elif four:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 4, "🙂")
        elif five:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 5, "😀")
        else:
            return None

    @staticmethod
    def _return_quantitative_ui_button(ui_type, disable_on_click, button_states, index, output):
        if disable_on_click:
            if st.session_state[f"{ui_type}_state_clicked"] == button_states[index - 1]:
                return output
            else:
                return None
        else:
            return output

    def _emoji_button(self, emoji, key, disable_on_click, button_states, index):
        return st.button(
            emoji,
            key=f"{key}_{index}",
            on_click=self._disable_buttons,
            args=(key, disable_on_click, index, button_states),
            disabled=st.session_state.get(f"{key}_{index}_disable", False),
        )

    @staticmethod
    def _disable_buttons(key, disable_on_click, index, button_states):
        """Disable all buttons except the one that was clicked, and re-enable all buttons if re-clicked."""
        if disable_on_click:
            if any([st.session_state.get(button_state + "_disable", False) for button_state in button_states]):
                st.session_state[f"{key}_state_clicked"] = ""
                for button_state in button_states:
                    st.session_state[button_state + "_disable"] = False
            else:
                enabled = button_states.pop(index - 1)
                st.session_state[enabled + "_disable"] = False
                st.session_state[f"{key}_state_clicked"] = enabled
                for button_state in button_states:
                    st.session_state[button_state + "_disable"] = True

    @staticmethod
    def _enable_all_buttons(feedback_type):
        for key in st.session_state:
            if key.endswith("_disable") and key.startswith(feedback_type):
                st.session_state[key] = False
            if key == f"{feedback_type}_state_clicked":
                st.session_state[key] = ""



from trubrics.integrations.streamlit.collect import FeedbackCollector
from trubrics.integrations.streamlit.experiment import (
    explore_testing_data,
    generate_what_if_streamlit,
)

__all__ = [
    "FeedbackCollector",
    "explore_testing_data",
    "generate_what_if_streamlit",
]



class ValidationOutputError(Exception):
    """An exception signalling that the output from a validation function is invalid."""


class TrubricValidationError(Exception):
    """An exception signalling a validation in a Trubric run has failed."""


class PandasSchemaError(Exception):
    """An exception signalling that two DataFrames do not have the same schema (column names & types)."""


class UnknownValidationError(Exception):
    """An exception signalling that validation is not recognised by a given validator."""


class MissingConfigPathError(Exception):
    """An exception signalling that the trubrics config file path does not exist."""


class MissingTrubricRunFileError(Exception):
    """An exception signalling that a TrubricRun file does not exist."""


class EstimatorTypeError(Exception):
    """An exception signalling that the estimator is not of type 'classifier' or 'regressor'."""


class SklearnMetricTypeError(Exception):
    """An exception signalling that the metric provided does not refer to an sklearn default metric."""


class ModelPredictionError(Exception):
    """An exception signalling that the model can't predict on a given set."""


class EmptyDfError(Exception):
    """An exception signalling that the dataframe is empty."""



import pandas as pd
from mlflow import MlflowClient
from mlflow.models.evaluation import ModelEvaluator

from trubrics.context import DataContext
from trubrics.validations import Trubric
from trubrics.validations.run import TrubricRun


class TrubricsValidator(ModelEvaluator):
    def can_evaluate(self, model_type, **kwargs):
        return model_type in ["classifier", "regressor"]

    # pylint: disable=unused-argument
    def evaluate(self, *, model, dataset, run_id, evaluator_config, **kwargs):
        """
        Args:
            run_id: The ID of the MLflow Run to which to log results
            dataset: mlflow dataset. Is overwritten if a data_context is provided in evaluator_config
            evaluator_config: A dictionary of additional configurations for the trubrics evaluator. Contains:
                trubric_path: the path to the trubric file
                model: the model to validate
                tags: optional tags to add to the trubric
                failing_severity: optional severity level to raise the trubric. Default is "error"
                custom_scorers: optional dict of custom scorers for computing custom metrics
                custom_validator: optional custom validator
                slicing_functions: optional dict of slicing functions
                data_context: an optional data context to validate the model on
        """
        self.client = MlflowClient()
        self.run_id = run_id
        self.evaluator_config = evaluator_config

        X = dataset.features_data
        y = dataset.labels_data
        testing_data = pd.DataFrame(X, columns=dataset.feature_names).assign(target=y)
        data_context = self.evaluator_config.get("data_context") or DataContext(
            target="target", testing_data=testing_data
        )

        trubric_run_context = TrubricRun(
            data_context=data_context,
            model=self.evaluator_config["model"],
            model_name=model.metadata.artifact_path,
            model_version=model.metadata.model_uuid,
            trubric=Trubric.parse_file(self.evaluator_config["trubric_path"]),
            tags=self.evaluator_config.get("tags") or ["mlflow-run"],
            failing_severity=self.evaluator_config.get("failing_severity") or "error",
            custom_scorers=self.evaluator_config.get("custom_scorers"),
            slicing_functions=self.evaluator_config.get("slicing_functions"),
            custom_validator=self.evaluator_config.get("custom_validator"),
            metadata={
                "mlflow_run_id": self.run_id,
                "mlflow.username": self.client.get_run(self.run_id).data.tags["mlflow.user"],
            },
        )

        new_trubric = trubric_run_context.set_new_trubric()
        self.client.log_dict(self.run_id, dictionary=new_trubric.dict(), artifact_file="demo-trubric.json")
        new_trubric.raise_trubric_failure()



import streamlit as st
from trubrics_utils import trubrics_config, trubrics_successful_feedback

from trubrics.context import DataContext
from trubrics.example import get_titanic_data_and_model
from trubrics.example import titanic_config as tc
from trubrics.integrations.streamlit import (
    FeedbackCollector,
    generate_what_if_streamlit,
)


@st.cache_resource
def init_trubrics():
    _, test_df, model = get_titanic_data_and_model()

    data_context = DataContext(
        testing_data=test_df,
        target="Survived",
        categorical_columns=tc.CATEGORICAL_COLUMNS,
        business_columns=tc.BUSINESS_COLUMNS,
    )

    return model, data_context


if "wi_prediction" not in st.session_state:
    st.session_state["wi_prediction"] = None

model, data_context = init_trubrics()
st.title("Titanic Demo App")
with st.sidebar:
    email, password, feedback_component, feedback_type = trubrics_config(default_component=False)
    with st.form("form"):
        st.subheader("Test the model with different inputs")
        df = generate_what_if_streamlit(data_context=data_context)
        submit = st.form_submit_button("Predict")

metadata = None
if submit:
    st.session_state["wi_prediction"] = model.predict(df)[0]
st.markdown("")

if st.session_state["wi_prediction"] is not None:
    metadata = {"data": df.to_dict(), "prediction": st.session_state["wi_prediction"]}

    prediction = (
        "## :orange["
        + str(st.session_state["wi_prediction"])
        + f"{' (survived)' if st.session_state['wi_prediction'] else  ' (did not survive)'}]"
    )
    col1, col2 = st.columns(2)
    with col1:
        st.markdown("## :orange[Model prediction:]")
    with col2:
        st.markdown(prediction, unsafe_allow_html=True)
    if email and password:
        collector = FeedbackCollector(
            component_name=feedback_component,
            email=email,
            password=password,
        )

        feedback = collector.st_feedback(
            feedback_type=feedback_type,
            model="your_model_name",
            open_feedback_label="[Optional] Provide additional feedback",
            metadata=metadata,
        )
        if feedback:
            trubrics_successful_feedback(feedback)
else:
    st.warning("Click 'Predict' in the sidebar to generate predictions.")



import streamlit as st


def trubrics_config(default_component: bool = True):
    st.subheader("Input your Trubrics credentials:")
    email = st.text_input(
        label="email", placeholder="email", label_visibility="collapsed", value=st.secrets.get("TRUBRICS_EMAIL", "")
    )

    password = st.text_input(
        label="password",
        placeholder="password",
        label_visibility="collapsed",
        type="password",
        value=st.secrets.get("TRUBRICS_PASSWORD", ""),
    )
    st.write("Don't have an account yet? Create one [here](https://trubrics.streamlit.app/)!")

    if default_component:
        return email, password

    feedback_component = st.text_input(
        label="feedback_component",
        placeholder="Feedback component name",
        label_visibility="collapsed",
    )

    feedback_type = st.radio(
        label="Select the component feedback type:", options=("faces", "thumbs", "textbox"), horizontal=True
    )

    return email, password, feedback_component, feedback_type


def trubrics_successful_feedback(feedback):
    st.write(
        ":green[You can now view your feedback] [here](https://trubrics.streamlit.app/), :green[and see the raw"
        " feedback response data here:]"
    )
    st.write(feedback)



TITLE = "Title"
TITLE_EXPLAIN = "Give the issue an explicit title."
DESCRIPTION = "Description"
DESCRIPTION_EXPLAIN = "Detail the issue you have observed."
FEEDBACK_SAVE_BUTTON = "Save feedback"
LOCAL_SAVE = "Feedback saved."
LOCAL_SAVE_HTML = f'<p style="color:Green;">{LOCAL_SAVE}</p>'
PLATFORM_SAVE = "Feedback saved to Trubrics."
PLATFORM_SAVE_HTML = f'<p style="color:Green;">{PLATFORM_SAVE}</p>'
FEEDBACK_NOT_SAVED = "Please specify both a feedback title and a description."
FEEDBACK_NOT_SAVED_HTML = '<p style="color:Red;">Please specify a feedback title and a description.</p>'
USER_EMAIL = "User email"
USER_PASSWORD = "User password"



import pandas as pd
import streamlit as st

from trubrics.context import DataContext


def generate_what_if_streamlit(data_context: DataContext):
    """
    Generate a what-if tool based on the testing_data of the DataContext.
    """
    out_df = pd.DataFrame()
    wi_data = data_context.testing_data
    if data_context.features:
        for col in data_context.features:
            series = wi_data[col]
            if data_context.business_columns is not None and col in data_context.business_columns.keys():
                renamed_col = data_context.business_columns[col]
            else:
                renamed_col = col
            if data_context.categorical_columns is None:
                raise ValueError(
                    "A list of categorical_columns must be specified in the DataContext for the what-if functionality."
                )
            if col in data_context.categorical_columns:
                out_df[col] = [st.selectbox(label=renamed_col, options=tuple(series.dropna().unique()))]
            else:
                out_df[col] = [
                    st.slider(
                        renamed_col,
                        min_value=int(series.min()),
                        max_value=int(series.max()),
                        value=round(series.mean()),
                    )
                ]
    return out_df


def explore_testing_data(data_context, model):
    test_set = "Full test set"
    test_set_errors = "Test set errors"
    filtered_target = "Filtered prediction value"
    data_view = st.selectbox(
        label="Which data would you like to view?",
        options=[test_set, test_set_errors, filtered_target],
        label_visibility="collapsed",
    )
    if data_view == test_set:
        st.dataframe(data_context.renamed_testing_data)
    elif data_view == test_set_errors:
        st.dataframe(
            _filter_testing_data_errors(data_context=data_context, model=model).rename(data_context.business_columns)
        )
    elif data_view == filtered_target:
        target_split = st.radio(
            label="Prediction value",
            options=data_context.testing_data[data_context.target].unique(),
            label_visibility="collapsed",
        )
        st.dataframe(data_context.testing_data.loc[lambda x: x[data_context.target] == target_split])


def _filter_testing_data_errors(data_context, model):
    predict_col = f"{data_context.target}_predictions"
    assign_kwargs = {predict_col: model.predict(data_context.testing_data)}
    return data_context.testing_data.assign(**assign_kwargs).loc[lambda x: x[data_context.target] != x[predict_col], :]



import pytest
from pydantic import ValidationError

from trubrics.context import DataContext
from trubrics.exceptions import PandasSchemaError
from trubrics.validations.dataclass import Validation


def test_data_context_attributes(dummy_testing_data):
    business_columns = {"feature 2": "feature 3"}
    dc = DataContext(testing_data=dummy_testing_data, target="target", business_columns=business_columns)

    assert dc.name == "my_dataset"
    assert dc.version == "v0.0.1"
    assert dc.features == ["feature_1", "feature 2"]
    assert all(
        [
            a == b
            for a, b in zip(
                dc.renamed_testing_data.columns, dummy_testing_data.rename(columns=business_columns).columns
            )
        ]
    )


@pytest.mark.parametrize(
    "training_data_rename_cols,kwargs,error_type",
    [
        ({"feature_1": "feature_one"}, {"target": "target"}, PandasSchemaError),
        ({}, {"target": "wrong_target"}, KeyError),
        ({}, {"categorical_columns": ["feature"]}, KeyError),
        ({}, {"categorical_columns": ["target"]}, KeyError),
        ({}, {"business_columns": {"feature 1": "feature"}}, KeyError),
        ({}, {"features": ["feature"]}, PandasSchemaError),
    ],
)
def test_data_context_raises(dummy_testing_data, training_data_rename_cols, kwargs, error_type):
    training_data = dummy_testing_data.rename(columns=training_data_rename_cols)
    with pytest.raises(error_type):
        DataContext(testing_data=dummy_testing_data, training_data=training_data, **kwargs)


@pytest.mark.parametrize(
    "kwargs,error_type",
    [
        ({"passed": "wrong_outcome", "severity": "experiment"}, ValidationError),
        ({"passed": True, "severity": "wrong_severity"}, KeyError),
    ],
)
def test_validation_context_raises(kwargs, error_type):
    with pytest.raises(error_type):
        Validation(
            validation_type="validate_something",
            validation_kwargs={"a kwarg": None},
            explanation="some docstring",
            **kwargs
        )



import pandas as pd
import pkg_resources  # type: ignore
from sklearn.compose import ColumnTransformer
from sklearn.ensemble import RandomForestClassifier
from sklearn.impute import SimpleImputer
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder

import trubrics.example.titanic_config as tc


def read_test_data():
    stream = pkg_resources.resource_stream(__name__, "titanic_data.csv")
    return pd.read_csv(stream)


def get_titanic_data_and_model():
    df = read_test_data()
    preprocessed_train = df.pipe(preprocess)

    numerical_cols = [col for col in preprocessed_train.columns if col not in tc.CATEGORICAL_COLUMNS + [tc.TARGET]]

    training_pipeline = init_training(tc.CATEGORICAL_COLUMNS, numerical_cols)
    X_train, X_test, y_train, y_test = train_test_split(
        preprocessed_train[tc.CATEGORICAL_COLUMNS + numerical_cols],
        preprocessed_train[tc.TARGET],
        test_size=0.33,
        random_state=88,
    )
    rf_model = training_pipeline.fit(X_train, y_train)

    return X_train.assign(Survived=y_train), X_test.assign(Survived=y_test), rf_model


def _strip_title(row):
    return row.split(",")[1].split(".")[0].replace(" ", "")


def preprocess(df):
    return (
        df.assign(Title=lambda x: x["Name"].apply(_strip_title))
        .assign(Fare=lambda x: x["Fare"].round(2))
        .drop(["Ticket", "Cabin", "Name", "PassengerId"], axis=1)
    )


def init_training(categorical_features, numerical_features):
    numeric_pipeline = Pipeline(steps=[("impute", SimpleImputer(strategy="mean")), ("scale", MinMaxScaler())])

    categorical_pipeline = Pipeline(
        steps=[
            ("impute", SimpleImputer(strategy="most_frequent")),
            ("one-hot", OneHotEncoder(handle_unknown="ignore", sparse=False)),
        ]
    )

    processor = ColumnTransformer(
        transformers=[
            ("number", numeric_pipeline, numerical_features),
            ("category", categorical_pipeline, categorical_features),
        ]
    )

    pipeline = Pipeline(steps=[("processor", processor), ("regressor", RandomForestClassifier(random_state=88))])
    return pipeline



from typing import Any, Dict, List, Optional

import gradio as gr

from trubrics.context import DataContext
from trubrics.feedback import config
from trubrics.feedback.dataclass import Feedback


def collect_feedback(
    path: Optional[str] = None,
    data_context: Optional[DataContext] = None,
    model_name: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
    save_ui: bool = False,
):
    """
    A component to collect user feedback within a Gradio web application.

    Args:
        path: path to save feedback local .json. Defaults to "./<timestamp>_feedback.json"
        data_context: the DataContext containing your datasets
        model_name: optional name of your model
        model_version: optional version of your model
        metadata: any metric which the user wants to save into the feedback issue such as
                  feature values, prediction, etc. Defaults to None.
        tags: list of any tags for the feedback issue. Defaults to None.
        save_ui: save to the Trubrics platform
    """

    def get_feedback(title: str, description: str, email: str, password: str):
        if not (len(title) == 0 or len(description) == 0):
            feedback = Feedback(
                feedback_type="issue",
                user_response={title: description},
                data=data_context.name if data_context else None,
                model=model_name,
                metadata=metadata,
                tags=tags,
            )
            if save_ui:
                feedback.save_ui(email, password)
                return config.PLATFORM_SAVE_HTML
            else:
                feedback.save_local(path=path)
                return config.LOCAL_SAVE_HTML
        else:
            return config.FEEDBACK_NOT_SAVED_HTML

    title = gr.Textbox(label=config.TITLE, placeholder=config.TITLE_EXPLAIN)
    description = gr.Textbox(label=config.DESCRIPTION, placeholder=config.DESCRIPTION_EXPLAIN, lines=5)
    if save_ui:
        with gr.Row():
            email = gr.Textbox(placeholder=config.USER_EMAIL, show_label=False, type="email")
            password = gr.Textbox(placeholder=config.USER_PASSWORD, show_label=False, type="password")
        inputs = [title, description, email, password]
    else:
        inputs = [title, description]
    feedback_button = gr.Button(config.FEEDBACK_SAVE_BUTTON)
    kwargs = {"fn": get_feedback, "inputs": inputs, "outputs": gr.HTML()}
    return feedback_button.click(**kwargs)



import openai
import streamlit as st
from trubrics_utils import trubrics_config, trubrics_successful_feedback

if "response" not in st.session_state:
    st.session_state["response"] = ""


st.title("LLM User Feedback with Trubrics")

with st.sidebar:
    email, password = trubrics_config()

models = ("text-davinci-003", "text-davinci-002")
model = st.selectbox(
    "Choose your GPT-3.5 LLM",
    models,
    help="Consult https://platform.openai.com/docs/models/gpt-3-5 for model info.",
)

openai.api_key = st.secrets.get("OPENAI_API_KEY")
if openai.api_key is None:
    raise ValueError("OpenAI key is missing. Set OPENAI_API_KEY in st.secrets")

prompt = st.text_area(label="Prompt", label_visibility="collapsed", placeholder="What would you like to know?")
button = st.button(f"Ask {model}")

if button:
    st.session_state["response"] = openai.Completion.create(model=model, prompt=prompt, temperature=0.5, max_tokens=200)

if st.session_state["response"]:
    response_text = st.session_state["response"].choices[0].text.replace("\n", "")
    st.markdown(f"#### :violet[{response_text}]")

    from trubrics.integrations.streamlit import FeedbackCollector

    if email and password:
        collector = FeedbackCollector(
            component_name="default",
            email=email,
            password=password,
        )

        feedback = collector.st_feedback(
            feedback_type="thumbs",
            model=model,
            open_feedback_label="[Optional] Provide additional feedback",
            metadata={"response": st.session_state["response"], "prompt": prompt},
        )

        if feedback:
            trubrics_successful_feedback(feedback)

    else:
        st.warning("To save some feedback to Trubrics, add your account details in the sidebar.")



from typing import Any, Dict, Iterator, List, Optional

from pydantic import BaseModel, validator
from rich import print as rprint
from sklearn.metrics._scorer import _BaseScorer

from trubrics.context import DataContext
from trubrics.exceptions import UnknownValidationError
from trubrics.validations import ModelValidator, Trubric, Validation


class TrubricRun(BaseModel):
    """The TrubricRun object to group all necessary code for a run. Load data and models from
    remote locations or locally for validation within a pipeline.

    Attributes:
        data_context: a data context to validate a model on
        model: a model to validate
        model_name: the name of the new model
        model_version: the version of the new model
        trubric: a Trubric object listing all validations to execute
        metadata: any new metadata to input to the Trubric
        tags: any new tags for the trubric
        custom_validator: an optional custom validator
        custom_scorers: an optional dict of custom scorers for computing custom metrics
        slicing_functions: an optional dict of slicing functions
    """

    data_context: DataContext
    model: Optional[Any]
    model_name: Optional[str]
    model_version: Optional[str]
    trubric: Trubric
    metadata: Optional[Dict[str, str]] = None
    tags: List[Optional[str]] = []
    custom_validator: Any = None
    custom_scorers: Optional[Dict[str, Any]] = None
    slicing_functions: Optional[Dict[str, Any]] = None
    failing_severity: Optional[str] = None

    @validator("custom_validator")
    def custom_validator_inherits_validator(cls, val):
        if val:
            if issubclass(val, ModelValidator):
                return val
            raise TypeError("Wrong type for 'custom_validator', must be subclass of ModelValidator.")

    @validator("custom_scorers")
    def custom_scorer_is_make_scorer(cls, val):
        if val:
            for scorer in val:
                if not issubclass(type(val[scorer]), _BaseScorer):
                    raise TypeError("Each scorer must be subclass of scikit-learn's _BaseScorer.")
            return val

    def generate_validations_from_trubric(self) -> Iterator[Validation]:
        if self.custom_validator is not None:
            model_validator = self.custom_validator(
                data=self.data_context,
                model=self.model,
                custom_scorers=self.custom_scorers,
                slicing_functions=self.slicing_functions,
            )
        else:
            model_validator = ModelValidator(
                data=self.data_context,
                model=self.model,
                custom_scorers=self.custom_scorers,
                slicing_functions=self.slicing_functions,
            )
        for validation in self.trubric.validations:
            args = validation.validation_kwargs["args"]
            kwargs = validation.validation_kwargs["kwargs"]
            try:
                validation_result = getattr(model_validator, validation.validation_type)(*args, **kwargs)
                new_validation = validation.copy()
                new_validation.passed = validation_result.passed
                new_validation.result = validation_result.result
                yield new_validation
            except AttributeError:
                raise UnknownValidationError(
                    f"The validation '{validation.validation_type}' does not appear to belong to a validator."
                    " Try adding the object that generated the validation to the 'custom_validator' parameter."
                )

    def set_new_trubric(self) -> Trubric:
        all_validation_results = self.generate_validations_from_trubric()
        validations = []
        for validation_result in all_validation_results:
            validations.append(validation_result)

            message_start = f"{validation_result.validation_type} [{validation_result.severity.upper()}]"
            completed_dots = f"[grey82]{(100 - len(message_start)) * '.'}[grey82]"
            message_end = (
                "[bold"
                f" {'green' if validation_result.passed else 'red'}]{'PASS' if validation_result.passed else 'FAIL'}"
            )
            rprint(message_start + completed_dots + message_end)

        return Trubric(
            name=self.trubric.name,
            failing_severity=self.failing_severity or self.trubric.failing_severity,
            model_name=self.model_name,
            model_version=self.model_version,
            data_context_name=self.data_context.name,
            data_context_version=self.data_context.version,
            metadata=self.metadata,
            tags=self.tags,
            validations=validations,
        ).set_dynamic_fields()



from typing import Any, Dict, List, Optional

import dash_bootstrap_components as dbc
from dash import Input, Output, callback, callback_context, html

from trubrics.context import DataContext
from trubrics.feedback import config
from trubrics.feedback.dataclass import Feedback


def collect_feedback(
    path: Optional[str] = None,
    data_context: Optional[DataContext] = None,
    model_name: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
    save_ui: bool = False,
):
    """
    A component to collect user feedback within a Dash web application.

    Args:
        path: path to save feedback local .json. Defaults to "./<timestamp>_feedback.json"
        data_context: the DataContext containing your datasets
        model_name: optional name of your model
        model_version: optional version of your model
        metadata: any metric which the user wants to save into the feedback issue such as
                  feature values, prediction, etc. Defaults to None.
        tags: list of any tags for the feedback issue. Defaults to None.
        save_ui: save to the Trubrics platform
    """
    title_input = html.Div(
        [
            dbc.Label(config.TITLE, html_for="title"),
            dbc.Input(id="title", placeholder=config.TITLE_EXPLAIN),
        ],
        className="mb-3",
    )

    description_input = html.Div(
        [
            dbc.Label(config.DESCRIPTION, html_for="description"),
            dbc.Input(id="description", placeholder=config.DESCRIPTION_EXPLAIN),
        ],
        className="mb-3",
    )

    button = html.Div(
        [
            html.Div(),
            dbc.Button(config.FEEDBACK_SAVE_BUTTON, id="button_input", color="secondary"),
            html.P(id="message"),
        ],
        className="mb-3",
    )

    @callback(
        Output("message", "children"),
        Output("message", "style"),
        Output("title", "value"),
        Output("description", "value"),
        Input("button_input", "n_clicks"),
        Input("title", "value"),
        Input("description", "value"),
    )
    def on_button_click(n, title, description):
        changed_id = [p["prop_id"] for p in callback_context.triggered][0]
        if "button_input" in changed_id:
            if title is None or description is None:
                return config.FEEDBACK_NOT_SAVED, {"color": "Red"}, title, description
            else:
                feedback = Feedback(
                    feedback_type="issue",
                    user_response={title: description},
                    data=data_context.name if data_context else None,
                    model=model_name,
                    metadata=metadata,
                    tags=tags,
                )
                feedback.save_local(path=path)
                if save_ui:
                    raise NotImplementedError()
                    # feedback.save_ui()
                return config.LOCAL_SAVE, {"color": "Green"}, None, None
        else:
            return None, None, title, description

    return dbc.Form([title_input, description_input, button])



from typing import Any, Dict, List, Optional

import pandas as pd
from loguru import logger
from pydantic import BaseModel, root_validator, validator

from trubrics.exceptions import (
    EstimatorTypeError,
    ModelPredictionError,
    PandasSchemaError,
)
from trubrics.utils.pandas import schema_is_equal


class DataContext(BaseModel):
    """
    The DataContext wraps ML datasets into a trubrics friendly format.

    Note:
        The DataContext *must contain* at least a testing_data and a target attribute.
        Default values are set for all other attributes.

    Attributes:
        name: DataContext name. Required for trubrics UI tracking.
        version: DataContext version. Required for trubrics UI tracking.
        testing_data: Dataframe that all validations are executed against. Should contain all features and target
                      values that the model was trained with.
        target: Name of target column.
        training_data: Dataframe of the training data.
        minimum_functionality_data: Dataframe of the minimum functionality of a model. This contains samples that
                                    the model should never fail on.
        features: list of model features. If None, all columns except the target column are set as features.
        categorical_columns: List of categorical names of the train & test datasets.
        business_columns: Mapping between dataset column names and comprehensible column names to be displayed to users.
    """

    name: str = "my_dataset"
    version: str = "v0.0.1"
    testing_data: pd.DataFrame
    target: str
    training_data: Optional[pd.DataFrame] = None
    minimum_functionality_data: Optional[pd.DataFrame] = None
    features: Optional[List[str]] = None
    categorical_columns: Optional[List[str]] = None
    business_columns: Optional[Dict[str, str]] = None

    class Config:
        allow_mutation = False
        arbitrary_types_allowed = True
        extra = "forbid"

    @property
    def X_test(self) -> pd.DataFrame:
        """Feature testing dataframe."""
        return self.testing_data[self.features]

    @property
    def y_test(self) -> pd.Series:
        """Target testing series."""
        return self.testing_data[self.target]

    @property
    def X_train(self) -> Optional[pd.DataFrame]:
        """Feature training dataframe."""
        return self.training_data[self.features] if self.training_data is not None else None

    @property
    def y_train(self) -> Optional[pd.Series]:
        """Target training series."""
        return self.training_data[self.target] if self.training_data is not None else None

    @property
    def renamed_testing_data(self) -> pd.DataFrame:
        """Renamed testing data with business columns"""
        if self.business_columns is None:
            raise TypeError("Business columns must be set to rename testing features.")
        return self.testing_data.rename(columns=self.business_columns)

    @root_validator
    def validate_features(cls, values) -> List[str]:
        """Features are here defined as all testing column names excluding the target feature.

        Note:
            @root_validator is used here to allow for @properties to use self.features.
        """
        v = values["features"]
        if v is None:
            values["features"] = [col for col in values["testing_data"].columns if col != values["target"]]
        elif set(v).issubset(set(values["testing_data"].columns)):
            values["features"] = v
        else:
            raise PandasSchemaError("All features must be present in testing_data.")
        return values

    @validator("target")
    def target_column_must_be_in_data(cls, v, values):
        if v not in values["testing_data"].columns:
            raise KeyError("Target column must be in testing_data column names.")
        return v

    @validator("training_data")
    def training_and_testing_must_have_same_schema(cls, v, values):
        if v is not None and not schema_is_equal(values["testing_data"], v):
            raise PandasSchemaError("Training and testing data must have identical schemas.")
        return v

    @validator("minimum_functionality_data")
    def minimum_functionality_and_testing_must_have_same_schema(cls, v, values):
        if v is not None and not schema_is_equal(values["testing_data"], v):
            raise PandasSchemaError("Minimum functionality and testing data must have identical schemas.")
        return v

    @validator("categorical_columns")
    def categorical_columns_must_be_in_data(cls, v, values):
        if not set(v).issubset(values["testing_data"].columns):
            raise KeyError("All categorical columns must be in testing_data column names.")
        return v

    @validator("categorical_columns")
    def target_column_must_not_be_in_categoricals(cls, v, values):
        if v is not None and values["target"] in v:
            raise ValueError(
                "Target column should not feature as a categorical column. Categorical columns only refer to features."
            )
        return v

    @validator("business_columns")
    def business_columns_must_be_in_data(cls, v, values):
        if not set(v.keys()).issubset(values["testing_data"].columns):
            raise KeyError("All business columns must be in testing_data column names.")
        return v


class TrubricsModel(BaseModel):
    data: DataContext
    model: Any

    class Config:
        allow_mutation = False
        arbitrary_types_allowed = True
        extra = "forbid"

    @validator("model")
    def does_model_predict_train_head(cls, v: Any, values: Any) -> Any:
        """
        Validate that model predicts on the first 5 rows of the training data.
        """
        if values["data"].training_data is not None:
            try:
                v.predict(values["data"].X_train.head())
            except ValueError:
                raise ModelPredictionError("The model specified does not predict on the train data.")
        return v

    @validator("model")
    def does_model_predict_test_head(cls, v: Any, values: Any) -> Any:
        """
        Validate that model predicts on the first 5 rows of the testing data.
        """
        try:
            v.predict(values["data"].X_test.head())
        except ValueError:
            raise ModelPredictionError("The model specified does not predict on the test data.")
        return v

    @property
    def model_type(self) -> Optional[str]:
        if self.model._estimator_type in ["regressor", "classifier"]:
            return self.model._estimator_type
        else:
            raise EstimatorTypeError("_estimator_type must be a 'regressor' or a 'classifier'.")

    @property
    def predictions_train(self) -> Optional[pd.Series]:
        if self.data.training_data is not None:
            logger.debug("Predicting train set.")
            return self.model.predict(self.data.X_train)
        else:
            raise ValueError("Training data not specified in DataContext.")

    @property
    def predictions_minimum_functionality(self) -> Optional[pd.Series]:
        if self.data.minimum_functionality_data is not None:
            logger.debug("Predicting minimum functionality set.")
            return self.model.predict(self.data.minimum_functionality_data[self.data.features])
        else:
            raise ValueError("Minimum functionality data not specified in DataContext.")

    @property
    def predictions_test(self) -> pd.Series:
        logger.debug("Predicting test set.")
        return self.model.predict(self.data.X_test)

    @property
    def probabilities_train(self) -> Optional[Dict[str, pd.Series]]:
        if self.data.training_data is not None:
            logger.debug("Predicting probabilities on train set.")
            probabilities = {}
            for _class, _proba in zip(self.model.classes_, self.model.predict_proba(self.data.X_train).T):
                probabilities[_class] = _proba
            return probabilities
        else:
            raise ValueError("Training data not specified in DataContext.")

    @property
    def probabilities_test(self) -> Dict[str, pd.Series]:
        logger.debug("Predicting probabilities on test set.")
        probabilities = {}
        for _class, _proba in zip(self.model.classes_, self.model.predict_proba(self.data.X_test).T):
            probabilities[_class] = _proba
        return probabilities



from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

from git import InvalidGitRepositoryError
from git.repo import Repo
from loguru import logger
from pydantic import BaseModel, validator

from trubrics.exceptions import TrubricValidationError
from trubrics.trubrics_platform.auth import (
    expire_after_n_seconds,
    get_trubrics_auth_token,
)
from trubrics.trubrics_platform.firestore import add_document_to_project_subcollection
from trubrics.trubrics_platform.trubrics_config import load_trubrics_config


def _validation_context_example():
    return {
        "example": {
            "validation_type": "validate_performance_against_threshold",
            "validation_kwargs": {"args": [], "kwargs": {"threshold": 0.8}},
            "passed": True,
            "severity": "error",
            "result": {"performance": "0.79"},
        }
    }


class Validation(BaseModel):
    """
    Dataclass for a single validation point. Must be serialisable to .json, as is fed into Trubric dataclass.

    Note:
        A Validation object constrains the output of validations, with the @validation_output decorator.

    Attributes:
        validation_type: method name of the validation.
        validation_kwargs: all args and kwargs that the validation had run with.
        passed: pass or fail output of the validation.
        severity: severity of the validation, can be one of ["error", "warning", "experiment"], is "error" by default
        result: a dictionary of contextual elements calculated during the validation run
        explanation: docstring explanation of the validation.
    """

    validation_type: str
    validation_kwargs: Dict[str, Optional[Any]]
    passed: bool
    severity: str = "error"
    result: Optional[Dict[str, Optional[Any]]]
    explanation: str

    class Config:
        extra = "forbid"
        validate_assignment = True
        schema_extra = _validation_context_example()

    @validator("severity")
    def severity_must_be(cls, v: str):
        severity_values = ["error", "warning", "experiment"]
        if v not in severity_values:
            raise KeyError(f"Severity must be set to: {severity_values}.")
        return v


class Trubric(BaseModel):
    """
    Dataclass for a trubric, or set of validation points. Must be serialisable to .json.
    A Trubric must contain at least metadata about the DataContext used to create the validations,
    and should contain metadata also about any model that is used.

    Attributes:
        name: trubric name
        passed: has trubric passed all validations (depends on the failing_severity)
        total_passed: number of validations that passed (depends on the failing_severity)
        total_failed: number of validations that failed (depends on the failing_severity)
        failing_severity: minimum severity that the trubric fails on, can be one of ["error", "warning", "experiment"].
        data_context_name: data context name (from DataContext)
        data_context_version: data context version (from DataContext)
        model_name: model name
        model_version: model version
        tags: list of tags for the trubric
        run_by: who the trubric was run by
        git_commit: a git commit hash from the git repo where the trubric was run
        timestamp: timestamp at which the trubric was run
        metadata: free textual metadata field
        validations: list of validations (defined by Validation)
    """

    name: str
    passed: Optional[bool] = None
    total_passed: Optional[int] = None
    total_failed: Optional[int] = None
    failing_severity: str = "error"
    data_context_name: str
    data_context_version: str
    model_name: Optional[str] = None
    model_version: Optional[str] = None
    tags: List[Optional[str]] = []
    run_by: Optional[Dict[str, str]] = None
    git_commit: Optional[str] = None
    timestamp: Optional[int] = None
    metadata: Optional[Dict[str, str]] = None
    validations: List[Validation]

    class Config:
        extra = "forbid"

    @validator("failing_severity")
    def failing_severity_must_be(cls, v: str):
        severity_values = ["error", "warning", "experiment"]
        if v not in severity_values:
            raise KeyError(f"Failing severity must be set to: {severity_values}.")
        return v

    def save_local(self, path: Optional[str] = None, raise_on_failure: bool = False):
        self.set_dynamic_fields()
        if path is None:
            path = f"./{self.name}.json"
        with open(Path(path).absolute(), "w") as file:
            file.write(self.json(indent=4))
            logger.info(f"Trubric saved to {path}.")

        if raise_on_failure:
            self.raise_trubric_failure()

    def save_ui(self, raise_on_failure: bool = False):
        trubrics_config = load_trubrics_config()
        auth = get_trubrics_auth_token(
            trubrics_config.firebase_api_key,
            trubrics_config.email,
            trubrics_config.password.get_secret_value(),
            rerun=expire_after_n_seconds(),
        )

        self.run_by = {"email": trubrics_config.email, "displayName": trubrics_config.username}
        self.set_dynamic_fields()

        res = add_document_to_project_subcollection(
            auth,
            firestore_api_url=trubrics_config.firestore_api_url,
            project=trubrics_config.project,
            subcollection="trubrics",
            document_id=self.timestamp,
            document_dict=self.dict(),
        )
        if "error" in res:
            error_msg = f"Error in pushing trubric to the Trubrics UI: {res}"
            logger.error(error_msg)
            raise Exception(error_msg)
        else:
            logger.info("Trubric saved to the Trubrics UI.")

        if raise_on_failure:
            self.raise_trubric_failure()

    def set_dynamic_fields(self):
        if self.failing_severity == "warning":
            failing_severity = ["error", "warning"]
        elif self.failing_severity == "experiment":
            failing_severity = ["error", "warning", "experiment"]
        else:
            failing_severity = ["error"]
        validations = [validation for validation in self.validations if validation.severity in failing_severity]
        self.total_passed = len([a for a in validations if a.passed])
        self.total_failed = len(validations) - self.total_passed
        self.passed = True if self.total_failed == 0 else False
        self.timestamp = int(datetime.now().timestamp())
        try:
            self.git_commit = Repo(search_parent_directories=True).head.object.hexsha
        except InvalidGitRepositoryError:
            self.git_commit = None
            logger.warning(
                "Current directory is not a git repository. Run `trubrics run` inside a git repository to save the"
                " commit hash."
            )
        return self

    def raise_trubric_failure(self):
        if not self.passed and self.total_failed:
            raise TrubricValidationError(
                "Trubric has failed on"
                f" {self.total_failed} {'validations' if self.total_failed > 1 else 'validation'} with minimum"
                f" severity='{self.failing_severity}'."
            )



import os

from pydantic import BaseModel, SecretStr


class TrubricsDefaults(BaseModel):
    # no stress, this is not a secret api key
    firebase_api_key: str = "AIzaSyB6WPIVzMaRCnlL1ZmRosrQsbQYYagZARQ"
    firebase_project_id: str = "trubrics-streamlit"
    trubrics_url: str = "https://trubrics.streamlit.com/"
    demo_sign_up_url: str = "https://trubrics.com/sign-up/"


class TrubricsConfig(BaseModel):
    email: str
    password: SecretStr
    username: str
    firebase_api_key: str
    firestore_api_url: str

    class Config:
        json_encoders = {SecretStr: lambda v: v.get_secret_value() if v else None}

    def save(self):
        config_path = os.path.join(os.path.expanduser("~"), ".trubrics_config.json")
        with open(config_path, "w") as file:
            file.write(self.json(indent=4))


def load_trubrics_config() -> TrubricsConfig:
    config_path = os.path.join(os.path.expanduser("~"), ".trubrics_config.json")
    if os.path.exists(config_path):
        return TrubricsConfig.parse_file(config_path)
    else:
        raise FileNotFoundError("Trubrics config file not found. Run `trubrics init` to generate this file.")



from loguru import logger

from trubrics.feedback.dataclass import Feedback
from trubrics.trubrics_platform.auth import (
    TrubricsConfig,
    expire_after_n_seconds,
    get_trubrics_auth_token,
)
from trubrics.trubrics_platform.firestore import (
    list_components_in_organisation,
    record_feedback,
)


def save_to_trubrics(trubrics_config: TrubricsConfig, feedback: Feedback) -> dict:
    auth = get_trubrics_auth_token(
        trubrics_config.firebase_api_key,
        trubrics_config.email,
        trubrics_config.password.get_secret_value(),
        rerun=expire_after_n_seconds(),
    )
    components = list_components_in_organisation(firestore_api_url=trubrics_config.firestore_api_url, auth=auth)
    if feedback.component_name not in components:
        logger.error(
            f"Component '{feedback.component_name}' not found in organisation"
            f" '{trubrics_config.firestore_api_url.split('/')[-1]}'. Components currently available: {components}."
        )
        raise ValueError("Component doesn't exist.")
    res = record_feedback(
        auth,
        firestore_api_url=trubrics_config.firestore_api_url,
        document_dict=feedback.dict(),
    )
    if "error" in res:
        logger.error(res["error"])
    else:
        logger.info("Feedback response saved to Trubrics.")

    return res



import json
import time
from functools import lru_cache
from typing import Dict, Optional

import requests  # type: ignore
from loguru import logger

from trubrics.trubrics_platform.firestore import get_trubrics_firestore_api_url
from trubrics.trubrics_platform.trubrics_config import TrubricsConfig, TrubricsDefaults


def expire_after_n_seconds(seconds=600):
    """Return the same value within `seconds` time period."""
    return round(time.time() / seconds)


def reset_trubrics_password(firebase_api_key, email) -> Dict[str, str]:
    try:
        r = requests.post(
            f"https://identitytoolkit.googleapis.com/v1/accounts:sendOobCode?key={firebase_api_key}",
            headers={"Content-Type": "application/json"},
            data=json.dumps({"requestType": "PASSWORD_RESET", "email": email}),
            timeout=5000,
        )
        r.raise_for_status()
        auth_response = json.loads(r.text)
        logger.info(f"User password link for {email} has been sent.")
        return auth_response
    except requests.exceptions.RequestException as err:
        logger.error(f"Error sending rest password link for {email}: {str(err)}.")
        return {"error": str(err)}


def create_trubrics_account(firebase_api_key, email, password) -> Dict[str, str]:
    try:
        r = requests.post(
            f"https://identitytoolkit.googleapis.com/v1/accounts:signUp?key={firebase_api_key}",
            headers={"Content-Type": "application/json"},
            data=json.dumps({"email": email, "password": password}),
            timeout=5000,
        )
        r.raise_for_status()
        auth_response = json.loads(r.text)
        logger.info(f"User account {email} has been created.")
        return auth_response
    except requests.exceptions.RequestException as err:
        logger.error(f"Error creating account for {email}: {str(err)}.")
        return {"error": str(err)}


@lru_cache(maxsize=32)
def get_trubrics_auth_token(firebase_api_key, email, password, rerun=None) -> Dict[str, str]:
    del rerun  # this variable is just used to force a refresh of lru_cache
    try:
        r = requests.post(
            f"https://identitytoolkit.googleapis.com/v1/accounts:signInWithPassword?key={firebase_api_key}",
            headers={"Content-Type": "application/json"},
            data=json.dumps({"email": email, "password": password, "returnSecureToken": True}),
            timeout=5000,
        )
        r.raise_for_status()
        auth_response = json.loads(r.text)
        logger.info(f"User {email} has been authenticated.")
        return {
            "idToken": auth_response["idToken"],
            "email": auth_response["email"],
            "uid": auth_response["localId"],
            "displayName": auth_response["displayName"],
        }
    except requests.exceptions.RequestException as err:
        logger.error(f"Error authenticating {email}: {str(err)}.")
        return {"error": str(err)}


def init(
    email: str,
    password: str,
    firebase_api_key: Optional[str] = None,
    firebase_project_id: Optional[str] = None,
) -> TrubricsConfig:
    if firebase_api_key or firebase_project_id:
        if firebase_api_key and firebase_project_id:
            defaults = TrubricsDefaults(firebase_api_key=firebase_api_key, firebase_project_id=firebase_project_id)
        else:
            raise ValueError("Both API key and firebase_project_id are required to change project.")
    else:
        defaults = TrubricsDefaults()

    auth = get_trubrics_auth_token(defaults.firebase_api_key, email, password, rerun=expire_after_n_seconds())
    if "error" in auth:
        raise Exception(f"Error in login email '{email}' to the Trubrics UI: {auth['error']}")
    else:
        firestore_api_url = get_trubrics_firestore_api_url(auth, defaults.firebase_project_id)

    if "error" in auth:
        error_msg = f"Error in pushing feedback issue with email '{email}' to the Trubrics UI: {auth['error']}"
        logger.error(error_msg)
        raise Exception(error_msg)
    else:
        return TrubricsConfig(
            email=email,
            password=password,  # type: ignore
            username=auth["displayName"],
            firebase_api_key=defaults.firebase_api_key,
            firestore_api_url=firestore_api_url,
        )



TARGET = "Survived"
CATEGORICAL_COLUMNS = ["Sex", "Embarked", "Title"]
LOCAL_RAW_TRAIN_FILENAME = "trubrics/example/titanic_data.csv"
BUSINESS_COLUMNS = {
    "Pclass": "Ticket class",
    "SibSp": "Number of siblings / spouses aboard the Titanic",
    "Parch": "Number of parents / children aboard the Titanic",
    "Embarked": "Port of embarkation",
}



import timeit
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pandas as pd
import sklearn.metrics
from loguru import logger
from sklearn.dummy import DummyClassifier, DummyRegressor
from sklearn.inspection import permutation_importance

from trubrics.context import DataContext, TrubricsModel
from trubrics.exceptions import EmptyDfError, EstimatorTypeError, SklearnMetricTypeError
from trubrics.validations.validation_output import (
    validation_output,
    validation_output_type,
)


class ModelValidator:
    def __init__(
        self,
        data: DataContext,
        model: Any,
        custom_scorers: Optional[Dict[str, Any]] = None,
        slicing_functions: Optional[Dict[str, Any]] = None,
    ):
        self.tm = TrubricsModel(data=data, model=model)
        self.model_type = self.tm.model_type
        if self.tm.data.features is None:
            raise ValueError("Features can't be None")
        else:
            self.features = self.tm.data.features
        self.custom_scorers = custom_scorers
        self.slicing_functions = slicing_functions

        # dict of computed performances {"dataset_name": {"metric_name": value}}
        self.performances: Dict[str, Dict[str, float]] = defaultdict(dict)
        # dict of feature importances {"dataset_name": {"feature_name": mean_value}}
        self.feature_importances: Dict[str, dict] = defaultdict(dict)
        # dict of number of rows per data slice {"dataset_name": dataset_size}}
        self.data_slice_sizes: Dict[str, float] = defaultdict()

    @validation_output
    def validate_minimum_functionality_in_range(
        self, range_value: Union[int, float] = 0, range_inclusive: bool = True, severity: Optional[str] = None
    ):
        """**Minimum functionality validation for a range output.**

        Validates that a model correctly predicts all points in a given set of data, within a range of values.
        This dataset must be set in the `minimum_functionality_data` parameter of the DataContext.

        Example:
            ```py
            from trubrics.validations import ModelValidator
            model_validator = ModelValidator(data=data_context, model=model)
            model_validator.validate_minimum_functionality_in_range(
                range_value=0,
                range_inclusive=True
            )
            ```

        Args:
            range_value: a value that is added to and subtracted from the target value for a given prediction,
                         to create a range of possible values that the prediction should fall between.
            range_inclusive: make range inclusive (x <= prediction <= y) or exclusive (x <= prediction <= y)
            severity: severity of the validation. Can be either ['error', 'warning', 'experiment']. \
                      If None, defaults to 'error'.

        Returns:
            True for success, false otherwise. With a results dictionary giving all data points where \
            the model's prediction did not fall between the range given.
        """
        return self._validate_minimum_functionality_in_range(range_value, range_inclusive=range_inclusive)

    def _validate_minimum_functionality_in_range(
        self,
        range_value: Union[int, float] = 0,
        range_inclusive: bool = True,
    ) -> validation_output_type:
        if self.model_type == "classifier":
            raise EstimatorTypeError(
                "Validation may only be applied to regressor model types."
                " Try 'validate_minimum_functionality' validation for classifier model types."
            )

        minimum_functionality_df = self.tm.data.minimum_functionality_data
        if minimum_functionality_df is None:
            raise ValueError("Specify minimum_functionality_data attribute in DataContext.")

        def filter_predictions_not_in_range(minimum_functionality_df, range_value, range_inclusive):
            if range_inclusive:
                return minimum_functionality_df.loc[
                    lambda x: (x["predictions"] >= x[self.tm.data.target] + range_value)
                    | (x["predictions"] <= x[self.tm.data.target] - range_value),
                    :,
                ]
            else:
                return minimum_functionality_df.loc[
                    lambda x: (x["predictions"] > x[self.tm.data.target] + range_value)
                    | (x["predictions"] < x[self.tm.data.target] - range_value),
                    :,
                ]

        minimum_functionality_df["predictions"] = self.tm.predictions_minimum_functionality
        errors_df = filter_predictions_not_in_range(minimum_functionality_df, range_value, range_inclusive)
        return (
            len(errors_df) == 0,
            {"number_of_failures": len(errors_df), "errors_df": errors_df.to_dict()} if len(errors_df) != 0 else {},
        )

    @validation_output
    def validate_minimum_functionality(self, severity: Optional[str] = None):
        """**Minimum functionality validation.**

        Validates that a model correctly predicts all points in a given set of data. This dataset must be set
        in the `minimum_functionality_data` parameter of the DataContext.

        Example:
            ```py
            from trubrics.validations import ModelValidator
            model_validator = ModelValidator(data=data_context, model=model)
            model_validator.validate_minimum_functionality()
            ```

        Args:
            severity: severity of the validation. Can be either ['error', 'warning', 'experiment']. \
                      If None, defaults to 'error'.

        Returns:
            True for success, false otherwise. With a results dictionary giving all data points that were not \
            correctly predicted by the model.
        """
        return self._validate_minimum_functionality()

    def _validate_minimum_functionality(self) -> validation_output_type:
        if self.model_type == "regressor":
            raise EstimatorTypeError(
                "Validation may only be applied to classifier model types."
                " Try 'validate_minimum_functionality_in_range' validation for regressor model types."
            )
        minimum_functionality_df = self.tm.data.minimum_functionality_data
        if minimum_functionality_df is None:
            raise ValueError("Specify minimum_functionality_data attribute in DataContext.")
        minimum_functionality_df.loc[:, "predictions"] = self.tm.predictions_minimum_functionality
        errors_df = minimum_functionality_df.loc[lambda x: x[self.tm.data.target] != x["predictions"], :]
        return (
            len(errors_df) == 0,
            {"number_of_failures": len(errors_df), "errors_df": errors_df.to_dict()} if len(errors_df) != 0 else {},
        )

    @validation_output
    def validate_performance_against_threshold(
        self,
        metric: str,
        threshold: float,
        dataset: str = "testing_data",
        data_slice: Optional[str] = None,
        severity: Optional[str] = None,
    ):
        """**Performance validation versus a fixed threshold value.**

        Compares performance of a model on any of the datasets in the DataContext to a hard coded threshold value.

        Example:
            ```py
            from trubrics.validations import ModelValidator
            model_validator = ModelValidator(data=data_context, model=model)
            model_validator.validate_performance_against_threshold(
                metric="recall",
                threshold=0.8
            )
            ```

        Args:
            metric: performance metric name defined in sklearn (sklearn.metrics.SCORERS) or in a \
                    custom scorer fed in when initialising the ModelValidator object.
            threshold: the performance threshold that the model must attain.
            dataset: the name of a dataset from the DataContext {'testing_data', 'training_data'}.
            data_slice: the name of the data slice, specified in the slicing_functions parameter of ModelValidator.
            severity: severity of the validation. Can be either {'error', 'warning', 'experiment'}. \
                      If None, defaults to 'error'.

        Returns:
            True for success, false otherwise. With a results dictionary giving the actual model performance calculated.
        """
        return self._validate_performance_against_threshold(metric, threshold, dataset, data_slice)

    def _validate_performance_against_threshold(
        self, metric: str, threshold: float, dataset: str = "testing_data", data_slice: Optional[str] = None
    ) -> validation_output_type:
        performance = self._score_data_context(metric, dataset, data_slice)
        return bool(performance > threshold), {
            "performance": performance,
            "sample_size": self.data_slice_sizes[self._get_renamed_dataset(dataset, data_slice)],
        }

    @validation_output
    def validate_performance_std_across_slices(
        self,
        metric: str,
        data_slices: List[str],
        std_threshold: float,
        dataset: str = "testing_data",
        include_global_performance: bool = False,
        severity: Optional[str] = None,
    ):
        """**Performance validation comparing data slices.**

        Validates that a list of model performances on different data slices from a given dataset has a lower
        standard deviation than a given threshold value.

        Example:
            ```py
            from trubrics.validations import ModelValidator
            slicing_functions = {"female": lambda x: x[x["Sex"]=="female"], "male": lambda x: x[x["Sex"]=="male"]}
            model_validator = ModelValidator(data=data_context, model=model, slicing_functions=slicing_functions)
            model_validator.validate_performance_std_across_slices(
                metric="recall",
                dataset="training_data",
                data_slices=["male", "female"],
                std_threshold=0.05,
                include_global_performance=True
            )
            ```

        Args:
            metric: performance metric name defined in sklearn (sklearn.metrics.SCORERS) or in a \
                    custom scorer fed in when initialising the ModelValidator object.
            data_slices: a list of of data slices, specified in the slicing_functions parameter of ModelValidator.
            std_threshold: the standard deviation threshold that must be superior to the standard deviation of all \
                data slice performances.
            dataset: the name of a dataset from the DataContext {'testing_data', 'training_data'}.
            include_global_performance: whether or not to include the dataset global performance in the list.
            severity: severity of the validation. Can be either {'error', 'warning', 'experiment'}. \
                      If None, defaults to 'error'.
        """
        return self._validate_performance_std_across_slices(
            metric, data_slices, std_threshold, dataset, include_global_performance
        )

    def _validate_performance_std_across_slices(
        self,
        metric: str,
        data_slices: List[str],
        std_threshold: float,
        dataset: str = "testing_data",
        include_global_performance: bool = False,
    ) -> validation_output_type:
        output: Dict[str, Dict[str, Union[float, int]]] = {"performances": {}, "sample_sizes": {}}
        for data_slice in data_slices:
            output["performances"][f"{dataset}_{data_slice}"] = self._score_data_context(metric, dataset, data_slice)
            output["sample_sizes"][f"{dataset}_{data_slice}"] = self.data_slice_sizes[
                self._get_renamed_dataset(dataset, data_slice)
            ]
        if include_global_performance:
            output["performances"][dataset] = self._score_data_context(metric, dataset, data_slice=None)
            output["sample_sizes"][dataset] = self.data_slice_sizes[self._get_renamed_dataset(dataset, data_slice=None)]
        return np.std(list(output["performances"].values())) < std_threshold, output  # type: ignore

    @validation_output
    def validate_test_performance_against_dummy(
        self,
        metric: str,
        strategy: str = "most_frequent",
        dummy_kwargs: Optional[dict] = None,
        data_slice: Optional[str] = None,
        severity: Optional[str] = None,
    ):
        """**Performance validation of testing data versus a dummy baseline model.**

        Trains a DummyClassifier / DummyRegressor from \
        [sklearn](https://scikit-learn.org/stable/modules/classes.html?highlight=dummy#module-sklearn.dummy)\
        and compares performance against the model on the test set.

        Example:
            ```py
            from trubrics.validations import ModelValidator
            model_validator = ModelValidator(data=data_context, model=model)
            model_validator.validate_test_performance_against_dummy(
                metric="accuracy",
                strategy="stratified"
            )
            ```

        Args:
            metric: performance metric name defined in sklearn (sklearn.metrics.SCORERS) or in a \
                    custom scorer fed in when initialising the ModelValidator object.
            strategy: strategy of scikit-learns dummy model.
            dummy_kwargs: kwargs to be passed to dummy model.
            data_slice: the name of the data slice, specified in the slicing_functions parameter of ModelValidator.
            severity: severity of the validation. Can be either ['error', 'warning', 'experiment']. \
                      If None, defaults to 'error'.

        Returns:
            True for success, false otherwise. With a results dictionary giving the model's \
            actual performance on the test set and the dummy model's performance.
        """
        return self._validate_test_performance_against_dummy(metric, strategy, dummy_kwargs, data_slice)

    def _validate_test_performance_against_dummy(
        self, metric: str, strategy: str, dummy_kwargs: Optional[dict] = None, data_slice: Optional[str] = None
    ) -> validation_output_type:
        test_performance = self._score_data_context(metric, dataset="testing_data", data_slice=data_slice)
        scorer = self._scorer(metric)
        if self.tm.data.training_data is None:
            raise TypeError("In order to train dummy classifier, training_data must be set in the DataContext.")

        if self.model_type == "classifier":
            Dummy = DummyClassifier
        elif self.model_type == "regressor":
            Dummy = DummyRegressor
        else:
            raise NotImplementedError()
        if dummy_kwargs:
            dummy_model = Dummy(strategy=strategy, **dummy_kwargs)
        else:
            dummy_model = Dummy(strategy=strategy)

        if data_slice:
            X_train, y_train = self._slice_data_with_slicing_function("training_data", data_slice)
            X_test, y_test = self._slice_data_with_slicing_function("testing_data", data_slice)
            sample_size = self.data_slice_sizes[self._get_renamed_dataset("testing_data", data_slice)]
        else:
            X_train, y_train = self.tm.data.X_train, self.tm.data.y_train
            X_test, y_test = self.tm.data.X_test, self.tm.data.y_test
            sample_size = self.data_slice_sizes[self._get_renamed_dataset("testing_data", data_slice=None)]

        dummy_model.fit(X_train, y_train)
        dummy_performance = scorer(dummy_model, X_test, y_test)

        return test_performance > dummy_performance, {
            "dummy_performance": dummy_performance,
            "test_performance": test_performance,
            "sample_size": sample_size,
        }

    @validation_output
    def validate_performance_between_train_and_test(
        self,
        metric: str,
        threshold: Union[int, float],
        data_slice: Optional[str] = None,
        severity: Optional[str] = None,
    ):
        """**Performance validation comparing training and test data scores.**

        Scores the test set and the train set in the DataContext, and validates whether the test score is \
        inferior to but also within a certain range of the train score. Can be used to validate for overfitting
        on the training set.

        Example:
            ```py
            from trubrics.validations import ModelValidator
            model_validator = ModelValidator(data=data_context, model=model)
            model_validator.validate_performance_between_train_and_test(
                metric="recall",
                threshold=0.3
            )
            ```

        Args:
            metric: performance metric name defined in sklearn (sklearn.metrics.SCORERS) or in a \
                      custom scorer fed in when initialising the ModelValidator object.
            threshold: a positive value representing the maximum allowable difference between the train and \
                         test score.
            data_slice: the name of the data slice, specified in the slicing_functions parameter of ModelValidator.
            severity: severity of the validation. Can be either ['error', 'warning', 'experiment']. \
                      If None, defaults to 'error'.

        Returns:
            True for success, false otherwise. With a results dictionary giving the model's \
            performance on test and train sets.
        """
        return self._validate_performance_between_train_and_test(metric, threshold, data_slice)

    def _validate_performance_between_train_and_test(
        self,
        metric: str,
        threshold: Union[int, float],
        data_slice: Optional[str] = None,
    ) -> validation_output_type:
        test_score = self._score_data_context(metric, dataset="testing_data", data_slice=data_slice)
        test_sample_size = self.data_slice_sizes[self._get_renamed_dataset("testing_data", data_slice)]
        train_score = self._score_data_context(metric, dataset="training_data", data_slice=data_slice)
        train_sample_size = self.data_slice_sizes[self._get_renamed_dataset("training_data", data_slice)]

        passed = test_score < train_score and test_score >= train_score - threshold
        return passed, {
            "train_performance": train_score,
            "test_performance": test_score,
            "train_sample_size": train_sample_size,
            "test_sample_size": test_sample_size,
        }

    @validation_output
    def validate_inference_time(self, threshold: float, n_executions: int = 100, severity: Optional[str] = None):
        """**Model inference time validation.**

        Validate the model's inference time on a single data point from the test set.

        Example:
            ```py
            from trubrics.validations import ModelValidator
            model_validator = ModelValidator(data=data_context, model=model)
            model_validator.validate_inference_time(threshold=0.04, n_executions=100)
            ```

        Args:
            threshold: number of seconds that the model inference time should be inferior to.
            n_executions: number of executions of the `.predict()` method for a single data point. \
                The inference time will be the mean of each of the n_executions.

        Returns:
            True for success, false otherwise. With a results dictionary giving the model's \
            average inference time (in seconds).
        """
        return self._validate_inference_time(threshold, n_executions)

    def _validate_inference_time(self, threshold: float, n_executions: int = 100) -> validation_output_type:
        single_data_point = self.tm.data.X_test.iloc[[0]]
        inference_time = (
            timeit.timeit(lambda: self.tm.model.predict(single_data_point), number=n_executions) / n_executions
        )
        return inference_time < threshold, {"inference_time": inference_time}

    @validation_output
    def validate_feature_in_top_n_important_features(
        self,
        feature: str,
        top_n_features: int,
        dataset: str = "testing_data",
        permutation_kwargs: Optional[Dict[str, Any]] = None,
        severity: Optional[str] = None,
    ):
        """**Feature importance validation for top n features.**

        Validates that a given feature is in the top n most important features. For calculation of feature \
        importance we are using sklearn's permutation_importance.

        Example:
            ```py
            from trubrics.validations import ModelValidator
            model_validator = ModelValidator(data=data_context, model=model)
            model_validator.validate_feature_in_top_n_important_features(
                dataset="testing_data",
                feature="feature_a",
                top_n_features=2,
            )
            ```

        Args:
            feature: feature to assess.
            top_n_features: the number of most important features that the named feature must be ranked in. E.g. if
                            top_n_features=2, the feature must be within the top two most important features.
            dataset: the name of a dataset from the DataContext to calculate feature importance on \
                {'testing_data', 'training_data'}.
            permutation_kwargs: kwargs to pass into the sklearn.inspection.permutation_importance function.

        Returns:
            True for success, false otherwise. With a results dictionary giving the actual feature importance ranking.
        """
        return self._validate_feature_in_top_n_important_features(feature, top_n_features, dataset, permutation_kwargs)

    def _validate_feature_in_top_n_important_features(
        self,
        feature: str,
        top_n_features: int,
        dataset: str = "testing_data",
        permutation_kwargs: Optional[Dict[str, Any]] = None,
    ) -> validation_output_type:
        count = 0
        feature_importance = self._compute_permutation_feature_importance(
            dataset=dataset, permutation_kwargs=permutation_kwargs
        )
        importances_mean = feature_importance["importances_mean"]
        feature_index = self.features.index(feature)
        for importance in importances_mean:
            if importance > importances_mean[feature_index]:
                count += 1

        feature_importances_dict = {
            key: value for key, value in zip(self.features, feature_importance["importances_mean"])
        }
        return count < top_n_features, {
            "feature_importance_ranking": count,
            "feature_importance": feature_importances_dict,
        }

    @validation_output
    def validate_feature_importance_between_train_and_test(
        self, top_n_features: Optional[int] = None, permutation_kwargs: Optional[Dict[str, Any]] = None
    ):
        """**Permutation feature importance validation between train and test sets.**

        Validates that the ranking of top n features is the same for both test and train sets. For calculation of \
        feature importance we are using sklearn's permutation_importance.

        Example:
            ```py
            from trubrics.validations import ModelValidator
            model_validator = ModelValidator(data=data_context, model=model)
            model_validator.validate_feature_importance_between_train_and_test(
                top_n_features=1
            )
            ```

        Args:
            top_n_features: the number of most important features to consider for comparison between train and test \
                            sets. E.g. if top_n_features=2, the train and test sets must have the same 2 most \
                            important features, in the same order.
            permutation_kwargs: kwargs to pass into the sklearn.inspection.permutation_importance function.

        Returns:
            True for success, false otherwise. With a results dictionary giving the actual feature importance ranking.
        """
        return self._validate_feature_importance_between_train_and_test(top_n_features, permutation_kwargs)

    def _validate_feature_importance_between_train_and_test(
        self, top_n_features: Optional[int] = None, permutation_kwargs: Optional[Dict[str, Any]] = None
    ) -> validation_output_type:
        train_fi = self._compute_permutation_feature_importance(
            dataset="testing_data", permutation_kwargs=permutation_kwargs
        )
        test_fi = self._compute_permutation_feature_importance(
            dataset="training_data", permutation_kwargs=permutation_kwargs
        )
        ordered_train_fi = {value: key for key, value in sorted(zip(train_fi["importances_mean"], self.features))}
        ordered_test_fi = {value: key for key, value in sorted(zip(test_fi["importances_mean"], self.features))}

        if top_n_features:
            top_n_features = -top_n_features

        is_same_order = [feature for feature in ordered_train_fi][top_n_features:] == [
            feature for feature in ordered_test_fi
        ][top_n_features:]
        return is_same_order, {
            "training_feature_importance": ordered_train_fi,
            "testing_feature_importance": ordered_test_fi,
        }

    def _scorer(self, metric: str) -> Callable[[Any, pd.DataFrame, pd.Series], float]:
        if metric in sklearn.metrics.SCORERS:
            scorer = sklearn.metrics.SCORERS[metric]
        else:
            if self.custom_scorers is not None and metric in self.custom_scorers:
                scorer = self.custom_scorers[metric]
            else:
                raise SklearnMetricTypeError(
                    f"The metric '{metric}' is not part of scikit-learns scorers, nor is it defined as a custom scorer"
                    " in the custom_scorers attribute. Run `sklearn.metrics.SCORERS` to list default scorers, or input"
                    " your custom scorer to the ModelValidator."
                )
        return scorer

    def _slice_data_with_slicing_function(self, dataset: str, data_slice: str):
        df = getattr(self.tm.data, dataset)
        if self.slicing_functions:
            if data_slice in self.slicing_functions:
                sliced_data = self.slicing_functions[data_slice](df)
                if not isinstance(sliced_data, pd.DataFrame):
                    raise TypeError(
                        f"Slicing function '{data_slice}' returned type {type(sliced_data)} but must return"
                        " pd.DataFrame."
                    )
                if len(sliced_data) == 0:
                    raise EmptyDfError(f"Slice '{data_slice}' returned length 0 on the dataset: {dataset}.")
            else:
                raise ValueError(
                    f"Slice '{data_slice}' does not exist in the slicing_functions parameter of the ModelValidator."
                )
        else:
            raise TypeError(
                "In order to use data slices, add all slicing functions to the slicing_functions parameter of the"
                " ModelValidator."
            )
        return sliced_data.loc[:, self.features], sliced_data.loc[:, self.tm.data.target]

    def _score_data_context(self, metric: str, dataset: str, data_slice: Optional[str]) -> float:
        renamed_dataset = self._get_renamed_dataset(dataset, data_slice)
        previously_computed_performance = self.performances.get(renamed_dataset, {}).get(metric)
        if previously_computed_performance:
            return previously_computed_performance

        scorer = self._scorer(metric)
        if data_slice:
            X, y = self._slice_data_with_slicing_function(dataset, data_slice)
        elif data_slice is None and dataset == "testing_data":
            X, y = self.tm.data.X_test, self.tm.data.y_test
        elif data_slice is None and dataset == "training_data":
            if self.tm.data.X_train is None or self.tm.data.y_train is None:
                raise ValueError("Training data not specified in DataContext.")
            else:
                X, y = self.tm.data.X_train, self.tm.data.y_train
        else:
            raise ValueError(
                "Method reserved for testing on datasets within the DataContext: {'testing_data', 'training_data'}."
            )
        self.data_slice_sizes[renamed_dataset] = len(X)
        self.performances[renamed_dataset][metric] = scorer(self.tm.model, X, y)
        return self.performances[renamed_dataset][metric]

    @staticmethod
    def _get_renamed_dataset(dataset: str, data_slice: Optional[str]):
        return f"{dataset}_{data_slice}" if data_slice else dataset

    def _compute_permutation_feature_importance(
        self, dataset: str, permutation_kwargs: Optional[Dict[str, Any]]
    ) -> dict:
        previously_computed_importance = self.feature_importances.get(dataset)
        if previously_computed_importance:
            return previously_computed_importance

        kwargs = {"n_repeats": 10, "random_state": 88, "n_jobs": -1}
        if permutation_kwargs is not None:
            kwargs.update(permutation_kwargs)

        if dataset == "testing_data":
            X, y = self.tm.data.X_test, self.tm.data.y_test
        elif dataset == "training_data":
            if self.tm.data.X_train is None or self.tm.data.y_train is None:
                raise ValueError("Training data not specified in DataContext.")
            else:
                X, y = self.tm.data.X_train, self.tm.data.y_train
        else:
            raise ValueError(
                "Method reserved for testing on datasets within the DataContext: {'testing_data', 'training_data'}."
            )
        logger.debug(f"Computing permutation feature importance for {dataset} with {kwargs['n_repeats']} permutations.")
        self.feature_importances[dataset] = dict(permutation_importance(self.tm.model, X, y, **kwargs))
        return self.feature_importances[dataset]



import json
from functools import wraps
from typing import Any, Callable, Dict, Tuple, Union

import numpy as np
from typeguard import check_type

from trubrics.context import TrubricsModel
from trubrics.exceptions import ValidationOutputError
from trubrics.validations.dataclass import Validation

validation_output_type = Tuple[Union[bool, np.bool_], Dict[str, Union[list, dict, str, int, float, np.generic]]]


def validation_output(func: Callable) -> Callable:
    """Decorative function for validation point outputs."""

    @wraps(func)
    def inner(*args, **kwargs) -> Validation:
        output = func(*args, **kwargs)

        try:
            check_type("output", output, validation_output_type)
        except TypeError:
            raise ValidationOutputError(
                f"Each validation should return a bool and a result with type: {validation_output_type}"
            )

        passed, result = output

        if kwargs.get("severity"):
            severity = kwargs.pop("severity")
        else:
            severity = "error"

        if len(args) > 1:
            args = args[1:]  # ignore self arg for class methods
        else:
            args = ()

        typed_args = _correct_types_for_json(args)
        typed_kwargs = _correct_types_for_json(kwargs)
        typed_result = _correct_types_for_json(result)

        _is_jsonable(typed_args, raise_error=True)
        _is_jsonable(typed_kwargs, raise_error=True)
        _is_jsonable(typed_result, raise_error=True)

        if func.__doc__ is None:
            raise TypeError(
                "Validations require a docstring to explain the validation's purpose and required parameters."
            )
        stripped_docstring = "\n".join(" ".join(line.split()) for line in func.__doc__.split("\n"))

        return Validation(
            validation_type=func.__name__,
            validation_kwargs={"args": typed_args, "kwargs": typed_kwargs},
            explanation=stripped_docstring,
            passed=passed,
            severity=severity,
            result=typed_result,  # type: ignore
        )

    return inner


def _correct_types_for_json(value: Any):
    if isinstance(value, dict):
        return {_correct_types_for_json(key): _correct_types_for_json(val) for key, val in value.items()}
    elif isinstance(value, list):
        return [_correct_types_for_json(arg) for arg in value]
    elif isinstance(value, np.generic):
        return value.item()
    elif value != value:
        return None  # replace NaN with None for json serialisation
    else:
        return value


def _is_jsonable(obj: Any, raise_error: bool = False) -> bool:
    try:
        json.dumps(obj)
        return True
    except (OverflowError, TypeError) as e:
        if isinstance(obj, TrubricsModel):
            return False
        elif raise_error:
            raise e
        else:
            return False



import json

import pkg_resources  # type: ignore

from trubrics.context import DataContext
from trubrics.example.titanic import get_titanic_data_and_model
from trubrics.validations import Trubric
from trubrics.validations.run import TrubricRun

train_df, test_df, model = get_titanic_data_and_model()

data_context = DataContext(
    name="titanic_dataset",
    version="0.0.1",
    testing_data=test_df,
    minimum_functionality_data=test_df.iloc[:5],
    training_data=train_df,
    target="Survived",
)

# Note: This method of reading json is only necessary from within python package.
# Call `Trubric.parse_file("your_trubric.json")` to read in a trubric directly.
stream = pkg_resources.resource_stream(__name__, "my_first_trubric.json")
trubric = Trubric.parse_obj(json.load(stream))

RUN_CONTEXT = TrubricRun(
    data_context=data_context,
    model_name="titanic_model",
    model_version="0.0.1",
    model=model,
    tags=["cli-demo"],
    trubric=trubric,
    failing_severity="warning",
)



import importlib.util
import sys

from trubrics.validations.run import TrubricRun


def import_python_module(module_path: str):
    try:
        spec = importlib.util.spec_from_file_location("module.name", module_path)
        lib = importlib.util.module_from_spec(spec)  # type: ignore
        sys.modules["module.name"] = lib
        spec.loader.exec_module(lib)  # type: ignore
    except AttributeError:
        raise AttributeError(f"Trubrics config {module_path} is not a python module.")
    return lib


def validate_trubric_run_context(trubric_run_path: str):
    tc = import_python_module(module_path=trubric_run_path)

    if hasattr(tc, "RUN_CONTEXT"):
        if not isinstance(tc.RUN_CONTEXT, TrubricRun):
            raise TypeError("'RUN_CONTEXT' attribute must be of type TrubricRun.")
    else:
        raise AttributeError("Trubrics config python module must contain an attribute 'RUN_CONTEXT'.")
    return tc
